#include "src/version.h"

char const desc_copyight[] = {
    "(c) 2014-2015,2022 by Andreas Florath <andreas@florath.net>"};
char const desc_license[] = {"License GPLv2+: GNU GPL version 2 or later "
                             "<http://gnu.org/licenses/gpl.html>."};
